import logging
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
)

from typing_extensions import override

import opik.dict_utils as dict_utils
import opik.llm_usage as llm_usage
from opik.api_objects import span
from opik.decorator import arguments_helpers, base_track_decorator
from opik.types import LLMProvider

import litellm
import litellm.types.utils

LOGGER = logging.getLogger(__name__)

KWARGS_KEYS_TO_LOG_AS_INPUTS = [
    "messages",
    "functions",
    "function_call",
    "tools",
    "tool_choice",
]
RESPONSE_KEYS_TO_LOG_AS_OUTPUT = ["choices"]


class LiteLLMCompletionTrackDecorator(base_track_decorator.BaseTrackDecorator):
    """
    An implementation of BaseTrackDecorator designed specifically for tracking
    calls of LiteLLM's `completion` and `acompletion` functions.

    Besides special processing for input arguments and response content, it
    overrides _streams_handler() method to work correctly with
    LiteLLM CustomStreamWrapper objects.
    """

    @override
    def _start_span_inputs_preprocessor(
        self,
        func: Callable,
        track_options: arguments_helpers.TrackOptions,
        args: Tuple,
        kwargs: Dict[str, Any],
    ) -> arguments_helpers.StartSpanParameters:
        assert (
            kwargs is not None
        ), "Expected kwargs to be not None in litellm.completion(**kwargs) or litellm.acompletion(**kwargs)"

        name = track_options.name if track_options.name is not None else func.__name__

        metadata = track_options.metadata if track_options.metadata is not None else {}

        input, new_metadata = dict_utils.split_dict_by_keys(
            kwargs, keys=KWARGS_KEYS_TO_LOG_AS_INPUTS
        )
        metadata = dict_utils.deepmerge(metadata, new_metadata)
        metadata.update(
            {
                "created_from": "litellm",
            }
        )

        tags = ["litellm"]

        model_name = kwargs.get("model", "")
        provider = _get_provider_from_model(model_name)

        result = arguments_helpers.StartSpanParameters(
            name=name,
            input=input,
            type=track_options.type,
            tags=tags,
            metadata=metadata,
            project_name=track_options.project_name,
            model=model_name,
            provider=provider,
        )

        return result

    @override
    def _end_span_inputs_preprocessor(
        self,
        output: Any,
        capture_output: bool,
        current_span_data: span.SpanData,
    ) -> arguments_helpers.EndSpanParameters:
        assert isinstance(
            output,
            (
                litellm.types.utils.ModelResponse,
                dict,  # For dict responses
            ),
        ), f"Expected ModelResponse or dict, got {type(output)}"

        # Handle both ModelResponse objects and dict responses
        if hasattr(output, "model_dump"):
            result_dict = output.model_dump(mode="json")
        elif isinstance(output, dict):
            result_dict = output
        else:
            result_dict = dict(output)

        output_data, metadata = dict_utils.split_dict_by_keys(
            result_dict, RESPONSE_KEYS_TO_LOG_AS_OUTPUT
        )

        opik_usage = None
        if result_dict.get("usage") is not None:
            model_name = result_dict.get("model", "")
            provider = _get_provider_from_model(model_name)

            if provider is not None:
                opik_usage = llm_usage.try_build_opik_usage_or_log_error(
                    provider=provider,
                    usage=result_dict["usage"],
                    logger=LOGGER,
                    error_message="Failed to log token usage from litellm call",
                )
            else:
                opik_usage = llm_usage.build_opik_usage_from_unknown_provider(
                    usage=result_dict["usage"],
                )

        model = result_dict.get("model")

        result = arguments_helpers.EndSpanParameters(
            output=output_data,
            usage=opik_usage,
            metadata=metadata,
            model=model,
        )

        return result

    @override
    def _streams_handler(  # type: ignore
        self,
        output: Any,
        capture_output: bool,
        generations_aggregator: Optional[Callable[[List[Any]], Any]],
    ) -> Optional[Any]:
        # Streaming is not currently supported for LiteLLM

        NOT_A_STREAM = None
        return NOT_A_STREAM


def _get_provider_from_model(model_name: str) -> Optional[LLMProvider]:
    """
    Extract the actual provider from the model name using LiteLLM's built-in method.
    """
    try:
        provider_info = litellm.get_llm_provider(model_name)
        provider_name = provider_info[1] if len(provider_info) > 1 else None

        # Map LiteLLM provider names to our LLMProvider enum
        # Based on mapping from Java backend CostService.java to ensure consistency
        provider_mapping = {
            "openai": LLMProvider.OPENAI,
            "vertex_ai-language-models": LLMProvider.GOOGLE_VERTEXAI,
            "gemini": LLMProvider.GOOGLE_AI,
            "anthropic": LLMProvider.ANTHROPIC,
            "vertex_ai-anthropic_models": LLMProvider.ANTHROPIC_VERTEXAI,
            "bedrock": LLMProvider.BEDROCK,
            "bedrock_converse": LLMProvider.BEDROCK,
            "groq": LLMProvider.GROQ,
        }

        return provider_mapping.get(provider_name, None)
    except Exception:
        return None



import logging
from typing import Dict, Any, Tuple

from ... import _logging


LOGGER = logging.getLogger(__name__)
LANGGRAPH_OUTPUT_SIZE_THRESHOLD = 5000


def split_big_langgraph_outputs(
    outputs: Dict[str, Any],
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """
    Split large LangGraph outputs to extract messages for thread display.

    Returns:
        tuple: (filtered_output_for_display, additional_metadata_for_span)

    LangGraph agents often produce complex outputs with large internal state
    that breaks thread display. This extracts conversational messages for
    clean thread display while preserving the full state in metadata.
    """
    if not isinstance(outputs, dict):
        return outputs, {}

    langgraph_like_output = "messages" in outputs and len(outputs) > 1
    if langgraph_like_output:
        output_str = str(outputs)
        output_size = len(output_str)

        if output_size > LANGGRAPH_OUTPUT_SIZE_THRESHOLD:
            _logging.log_once_at_level(
                logging.WARNING,
                f"Filtering large LangGraph output ({output_size} chars) for thread display",
                LOGGER,
            )

            filtered_output = {
                "messages": outputs["messages"],
            }

            if "thread_id" in outputs:
                filtered_output["thread_id"] = outputs["thread_id"]

            additional_metadata = {
                "_opik_langgraph_full_output": outputs,
                "_opik_output_filtering": {
                    "filtered": True,
                    "original_size_chars": output_size,
                    "filtered_keys": [
                        k for k in outputs.keys() if k not in ["messages", "thread_id"]
                    ],
                    "reason": "Large LangGraph output filtered for better thread display",
                },
            }

            return filtered_output, additional_metadata

    return outputs, {}



# This litellm integration is currently not exposed in the documentation and
# used as a part of Opik CrewAI integration only.
#
# TODO: in order to support it officially - implement streaming support, update documentation, make it
# respect the OpikLogger callback in litellm repository to avoid confusion.

from .opik_tracker import track_litellm

__all__ = ["track_litellm"]



# This file was auto-generated by Fern from our API Definition.

import typing

from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.request_options import RequestOptions
from .raw_client import AsyncRawOpenTelemetryIngestionClient, RawOpenTelemetryIngestionClient


class OpenTelemetryIngestionClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawOpenTelemetryIngestionClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> RawOpenTelemetryIngestionClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawOpenTelemetryIngestionClient
        """
        return self._raw_client

    def receive_protobuf_traces(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> typing.Optional[typing.Any]:
        """
        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        typing.Optional[typing.Any]
            default response

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.open_telemetry_ingestion.receive_protobuf_traces()
        """
        _response = self._raw_client.receive_protobuf_traces(request_options=request_options)
        return _response.data


class AsyncOpenTelemetryIngestionClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawOpenTelemetryIngestionClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawOpenTelemetryIngestionClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawOpenTelemetryIngestionClient
        """
        return self._raw_client

    async def receive_protobuf_traces(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> typing.Optional[typing.Any]:
        """
        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        typing.Optional[typing.Any]
            default response

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.open_telemetry_ingestion.receive_protobuf_traces()
        asyncio.run(main())
        """
        _response = await self._raw_client.receive_protobuf_traces(request_options=request_options)
        return _response.data



import logging
import datetime
from typing import Any, Dict, List, Literal, Optional, Set, TYPE_CHECKING, cast, Tuple
import contextvars
from uuid import UUID

from langchain_core import language_models
from langchain_core.tracers import BaseTracer
from langchain_core.tracers.schemas import Run

import opik.dict_utils as dict_utils
import opik.llm_usage as llm_usage
from opik.api_objects import span, trace
from opik.types import DistributedTraceHeadersDict, ErrorInfoDict
from opik.validation import parameters_validator
from . import (
    base_llm_patcher,
    helpers as langchain_helpers,
    opik_encoder_extension,
    provider_usage_extractors,
)

from ...api_objects import helpers, opik_client
import opik.context_storage as context_storage
import opik.decorator.tracing_runtime_config as tracing_runtime_config

if TYPE_CHECKING:
    from langchain_core.runnables.graph import Graph
    from langchain_core.messages import BaseMessage

LOGGER = logging.getLogger(__name__)

opik_encoder_extension.register()

language_models.BaseLLM.dict = base_llm_patcher.base_llm_dict_patched()

SpanType = Literal["llm", "tool", "general"]


def _get_span_type(run: Dict[str, Any]) -> SpanType:
    if run.get("run_type") in ["llm", "tool"]:
        return cast(SpanType, run.get("run_type"))

    if run.get("run_type") in ["prompt"]:
        return cast(SpanType, "tool")

    return cast(SpanType, "general")


class OpikTracer(BaseTracer):
    """Langchain Opik Tracer.

    Args:
        tags: List of tags to be applied to each trace logged by the tracer.
        metadata: Additional metadata for each trace logged by the tracer.
        graph: A LangGraph Graph object to track the Graph Definition in Opik.
        project_name: The name of the project to log data.
    """

    def __init__(
        self,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        graph: Optional["Graph"] = None,
        project_name: Optional[str] = None,
        distributed_headers: Optional[DistributedTraceHeadersDict] = None,
        thread_id: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        validator = parameters_validator.create_validator(
            method_name="__init__", class_name=self.__class__.__name__
        )
        validator.add_str_parameter(thread_id, name="thread_id")
        validator.add_str_parameter(project_name, name="project_name")
        validator.add_dict_parameter(metadata, name="metadata")
        validator.add_list_parameter(tags, name="tags")
        if not validator.validate():
            validator.raise_validation_error()

        super().__init__(**kwargs)
        self._trace_default_metadata = metadata if metadata is not None else {}
        self._trace_default_metadata["created_from"] = "langchain"

        if graph:
            self._trace_default_metadata["_opik_graph_definition"] = {
                "format": "mermaid",
                "data": graph.draw_mermaid(),
            }

        self._trace_default_tags = tags

        self._span_data_map: Dict[UUID, span.SpanData] = {}
        """Map from run id to span data."""

        self._created_traces_data_map: Dict[UUID, trace.TraceData] = {}
        """Map from run id to trace data."""

        self._created_traces: List[trace.Trace] = []

        self._externally_created_traces_ids: Set[str] = set()

        self._project_name = project_name

        self._distributed_headers = distributed_headers

        self._opik_client = opik_client.get_client_cached()

        self._thread_id = thread_id

        self._opik_context_storage = context_storage.get_current_context_instance()

        self._root_run_external_parent_span_id: contextvars.ContextVar[
            Optional[str]
        ] = contextvars.ContextVar("root_run_external_parent_span_id", default=None)

    def _is_opik_span_created_by_this_tracer(self, span_id: str) -> bool:
        return any(span_.id == span_id for span_ in self._span_data_map.values())

    def _is_opik_trace_created_by_this_tracer(self, trace_id: str) -> bool:
        return any(
            trace_.id == trace_id for trace_ in self._created_traces_data_map.values()
        )

    def _persist_run(self, run: Run) -> None:
        if run.id not in self._span_data_map:
            LOGGER.warning(
                f"Span data for run '{run.id}' not found in the span data map. Skipping processing of _persist_run."
            )
            return

        span_data = self._span_data_map[run.id]

        run_dict: Dict[str, Any] = run.dict()

        error_info: Optional[ErrorInfoDict]
        trace_additional_metadata: Dict[str, Any] = {}

        if run_dict["error"] is not None:
            output = None
            error_info = ErrorInfoDict(
                exception_type="Exception",
                traceback=run_dict["error"],
            )
        else:
            output, trace_additional_metadata = (
                langchain_helpers.split_big_langgraph_outputs(run_dict["outputs"])
            )
            error_info = None

        if (
            span_data.parent_span_id is not None
            and self._is_opik_span_created_by_this_tracer(span_data.parent_span_id)
        ):
            # Langchain lost the parent-child relationship for Run, so it calls _persist_run
            # for a subchain when the ACTUAL root run has not yet persisted.
            # However, we know that this tracer created the parent span, so we don't
            # want to finalize the trace
            return

        self._ensure_no_hanging_opik_tracer_spans()

        if span_data.trace_id not in self._externally_created_traces_ids:
            trace_data = self._created_traces_data_map[run.id]

            # workaround for `.astream()` method usage
            if trace_data.input == {"input": ""}:
                trace_data.input = run_dict["inputs"]

            if trace_additional_metadata:
                trace_data.update(metadata=trace_additional_metadata)

            trace_data.init_end_time().update(output=output, error_info=error_info)
            trace_ = self._opik_client.trace(**trace_data.as_parameters)

            assert trace_ is not None
            self._created_traces.append(trace_)
            self._opik_context_storage.pop_trace_data(ensure_id=trace_data.id)

    def _ensure_no_hanging_opik_tracer_spans(self) -> None:
        root_run_external_parent_span_id = self._root_run_external_parent_span_id.get()
        there_were_no_external_spans_before_chain_invocation = (
            root_run_external_parent_span_id is None
        )

        if there_were_no_external_spans_before_chain_invocation:
            self._opik_context_storage.clear_spans()
        else:
            assert root_run_external_parent_span_id is not None
            self._opik_context_storage.trim_span_data_stack_to_certain_span(
                root_run_external_parent_span_id
            )

    def _track_root_run(
        self, run_dict: Dict[str, Any]
    ) -> Tuple[Optional[trace.TraceData], span.SpanData]:
        run_metadata = run_dict["extra"].get("metadata", {})
        root_metadata = dict_utils.deepmerge(self._trace_default_metadata, run_metadata)
        self._update_thread_id_from_metadata(run_dict)

        if self._distributed_headers:
            new_span_data = self._attach_span_to_distributed_headers(
                run_dict=run_dict,
                root_metadata=root_metadata,
            )
            return None, new_span_data

        current_span_data = self._opik_context_storage.top_span_data()
        self._root_run_external_parent_span_id.set(
            current_span_data.id if current_span_data is not None else None
        )
        if current_span_data is not None:
            new_span_data = self._attach_span_to_existing_span(
                run_dict=run_dict,
                current_span_data=current_span_data,
                root_metadata=root_metadata,
            )
            return None, new_span_data

        current_trace_data = self._opik_context_storage.get_trace_data()
        if current_trace_data is not None:
            new_span_data = self._attach_span_to_existing_trace(
                run_dict=run_dict,
                current_trace_data=current_trace_data,
                root_metadata=root_metadata,
            )
            return None, new_span_data

        return self._initialize_span_and_trace_from_scratch(
            run_dict=run_dict, root_metadata=root_metadata
        )

    def _initialize_span_and_trace_from_scratch(
        self, run_dict: Dict[str, Any], root_metadata: Dict[str, Any]
    ) -> Tuple[trace.TraceData, span.SpanData]:
        trace_data = trace.TraceData(
            name=run_dict["name"],
            input=run_dict["inputs"],
            metadata=root_metadata,
            tags=self._trace_default_tags,
            project_name=self._project_name,
            thread_id=self._thread_id,
        )
        span_data = span.SpanData(
            trace_id=trace_data.id,
            parent_span_id=None,
            name=run_dict["name"],
            input=run_dict["inputs"],
            type=_get_span_type(run_dict),
            metadata=root_metadata,
            tags=self._trace_default_tags,
            project_name=self._project_name,
        )
        return trace_data, span_data

    def _attach_span_to_existing_span(
        self,
        run_dict: Dict[str, Any],
        current_span_data: span.SpanData,
        root_metadata: Dict[str, Any],
    ) -> span.SpanData:
        project_name = helpers.resolve_child_span_project_name(
            current_span_data.project_name,
            self._project_name,
        )

        span_data = span.SpanData(
            trace_id=current_span_data.trace_id,
            parent_span_id=current_span_data.id,
            name=run_dict["name"],
            input=run_dict["inputs"],
            metadata=root_metadata,
            tags=self._trace_default_tags,
            project_name=project_name,
            type=_get_span_type(run_dict),
        )
        if not self._is_opik_trace_created_by_this_tracer(span_data.trace_id):
            self._externally_created_traces_ids.add(span_data.trace_id)

        return span_data

    def _attach_span_to_existing_trace(
        self,
        run_dict: Dict[str, Any],
        current_trace_data: trace.TraceData,
        root_metadata: Dict[str, Any],
    ) -> span.SpanData:
        project_name = helpers.resolve_child_span_project_name(
            current_trace_data.project_name,
            self._project_name,
        )

        span_data = span.SpanData(
            trace_id=current_trace_data.id,
            parent_span_id=None,
            name=run_dict["name"],
            input=run_dict["inputs"],
            metadata=root_metadata,
            tags=self._trace_default_tags,
            project_name=project_name,
            type=_get_span_type(run_dict),
        )
        if not self._is_opik_trace_created_by_this_tracer(current_trace_data.id):
            self._externally_created_traces_ids.add(current_trace_data.id)
        return span_data

    def _attach_span_to_distributed_headers(
        self,
        run_dict: Dict[str, Any],
        root_metadata: Dict[str, Any],
    ) -> span.SpanData:
        if self._distributed_headers is None:
            raise ValueError("Distributed headers are not set")

        span_data = span.SpanData(
            trace_id=self._distributed_headers["opik_trace_id"],
            parent_span_id=self._distributed_headers["opik_parent_span_id"],
            name=run_dict["name"],
            input=run_dict["inputs"],
            metadata=root_metadata,
            tags=self._trace_default_tags,
            project_name=self._project_name,
            type=_get_span_type(run_dict),
        )
        self._externally_created_traces_ids.add(span_data.trace_id)
        return span_data

    def _process_start_span(self, run: Run) -> None:
        try:
            self._process_start_span_unsafe(run)
        except Exception as e:
            LOGGER.error("Failed during _process_start_span: %s", e, exc_info=True)

    def _process_start_span_unsafe(self, run: Run) -> None:
        run_dict: Dict[str, Any] = run.dict()
        new_span_data: span.SpanData

        if not run.parent_run_id:
            # This is the first run for the chain.
            new_trace_data, new_span_data = self._track_root_run(run_dict)
            if new_trace_data is not None:
                self._opik_context_storage.set_trace_data(new_trace_data)
                if (
                    self._opik_client.config.log_start_trace_span
                    and tracing_runtime_config.is_tracing_active()
                ):
                    self._opik_client.trace(**new_trace_data.as_start_parameters)

            # save new span and trace data to local maps to be able to retrieve them later
            self._save_span_trace_data_to_local_maps(
                run_id=run.id,
                span_data=new_span_data,
                trace_data=new_trace_data,
            )

            self._opik_context_storage.add_span_data(new_span_data)
            if (
                self._opik_client.config.log_start_trace_span
                and tracing_runtime_config.is_tracing_active()
            ):
                self._opik_client.span(**new_span_data.as_start_parameters)
            return

        parent_span_data = self._span_data_map[run.parent_run_id]

        project_name = helpers.resolve_child_span_project_name(
            parent_span_data.project_name,
            self._project_name,
        )

        new_span_data = span.SpanData(
            trace_id=parent_span_data.trace_id,
            parent_span_id=parent_span_data.id,
            input=run_dict["inputs"],
            metadata=run_dict["extra"],
            name=run.name,
            type=_get_span_type(run_dict),
            project_name=project_name,
        )
        new_span_data.update(metadata={"created_from": "langchain"})

        self._save_span_trace_data_to_local_maps(
            run_id=run.id,
            span_data=new_span_data,
            trace_data=None,
        )

        if new_span_data.trace_id not in self._externally_created_traces_ids:
            self._created_traces_data_map[run.id] = self._created_traces_data_map[
                run.parent_run_id
            ]

        self._opik_context_storage.add_span_data(new_span_data)
        if (
            self._opik_client.config.log_start_trace_span
            and tracing_runtime_config.is_tracing_active()
        ):
            self._opik_client.span(**new_span_data.as_start_parameters)

    def _process_end_span(self, run: Run) -> None:
        span_data = None
        try:
            if run.id not in self._span_data_map:
                LOGGER.warning(
                    f"Span data for run '{run.id}' not found in the span data map. Skipping processing of end span."
                )
                return
            span_data = self._span_data_map[run.id]
            run_dict: Dict[str, Any] = run.dict()

            usage_info = provider_usage_extractors.try_extract_provider_usage_data(
                run_dict
            )
            if usage_info is None:
                usage_info = llm_usage.LLMUsageInfo()

            # workaround for `.astream()` method usage
            if span_data.input == {"input": ""}:
                span_data.input = run_dict["inputs"]

            filtered_output, additional_metadata = (
                langchain_helpers.split_big_langgraph_outputs(run_dict["outputs"])
            )

            if additional_metadata:
                span_data.update(metadata=additional_metadata)

            span_data.init_end_time().update(
                output=filtered_output,
                usage=(
                    usage_info.usage.provider_usage.model_dump()
                    if isinstance(usage_info.usage, llm_usage.OpikUsage)
                    else usage_info.usage
                ),
                provider=usage_info.provider,
                model=usage_info.model,
            )

            if tracing_runtime_config.is_tracing_active():
                self._opik_client.span(**span_data.as_parameters)
        except Exception as e:
            LOGGER.error(f"Failed during _process_end_span: {e}", exc_info=True)
        finally:
            if span_data is not None:
                self._opik_context_storage.trim_span_data_stack_to_certain_span(
                    span_id=span_data.id
                )
                self._opik_context_storage.pop_span_data(ensure_id=span_data.id)

    def _process_end_span_with_error(self, run: Run) -> None:
        if run.id not in self._span_data_map:
            LOGGER.warning(
                f"Span data for run '{run.id}' not found in the span data map. Skipping processing of _process_end_span_with_error."
            )
            return

        span_data = None
        try:
            run_dict: Dict[str, Any] = run.dict()
            span_data = self._span_data_map[run.id]
            error_info: ErrorInfoDict = {
                "exception_type": "Exception",
                "traceback": run_dict["error"],
            }

            span_data.init_end_time().update(
                output=None,
                error_info=error_info,
            )
            if tracing_runtime_config.is_tracing_active():
                self._opik_client.span(**span_data.as_parameters)
        except Exception as e:
            LOGGER.debug(f"Failed during _process_end_span_with_error: {e}")
        finally:
            if span_data is not None:
                self._opik_context_storage.trim_span_data_stack_to_certain_span(
                    span_id=span_data.id
                )
                self._opik_context_storage.pop_span_data(ensure_id=span_data.id)

    def _update_thread_id_from_metadata(self, run_dict: Dict[str, Any]) -> None:
        if not self._thread_id:
            # We want to default to any manually set thread_id, so only update if self._thread_id is not already set
            thread_id = run_dict["extra"].get("metadata", {}).get("thread_id")

            if thread_id:
                self._thread_id = thread_id

    def _save_span_trace_data_to_local_maps(
        self,
        run_id: UUID,
        span_data: Optional[span.SpanData],
        trace_data: Optional[trace.TraceData],
    ) -> None:
        if span_data is not None:
            self._span_data_map[run_id] = span_data

        if trace_data is not None:
            self._created_traces_data_map[run_id] = trace_data

    def flush(self) -> None:
        """
        Flush to ensure all data is sent to the Opik server.
        """
        self._opik_client.flush()

    def created_traces(self) -> List[trace.Trace]:
        """
        Get a list of traces created by OpikTracer.

        Returns:
            List[Trace]: A list of traces.
        """
        return self._created_traces

    def _skip_tracking(self) -> bool:
        if not tracing_runtime_config.is_tracing_active():
            return True

        return False

    def _on_llm_start(self, run: Run) -> None:
        """Process the LLM Run upon start."""
        if self._skip_tracking():
            return

        self._process_start_span(run)

    def on_chat_model_start(
        self,
        serialized: Dict[str, Any],
        messages: List[List["BaseMessage"]],
        *,
        run_id: UUID,
        tags: Optional[List[str]] = None,
        parent_run_id: Optional[UUID] = None,
        metadata: Optional[Dict[str, Any]] = None,
        name: Optional[str] = None,
        **kwargs: Any,
    ) -> Run:
        """Start a trace for an LLM run.

        Duplicated from Langchain tracer, it is disabled by default in all tracers, see https://github.com/langchain-ai/langchain/blob/fdda1aaea14b257845a19023e8af5e20140ec9fe/libs/core/langchain_core/callbacks/manager.py#L270-L289 and https://github.com/langchain-ai/langchain/blob/fdda1aaea14b257845a19023e8af5e20140ec9fe/libs/core/langchain_core/tracers/core.py#L168-L180

        Args:
            serialized: The serialized model.
            messages: The messages.
            run_id: The run ID.
            tags: The tags. Defaults to None.
            parent_run_id: The parent run ID. Defaults to None.
            metadata: The metadata. Defaults to None.
            name: The name. Defaults to None.
            kwargs: Additional keyword arguments.

        Returns:
            Run: The run.
        """
        start_time = datetime.datetime.now(datetime.timezone.utc)
        if metadata:
            kwargs.update({"metadata": metadata})

        # We switched from langchain dumpd to model_dump() as we don't need all the langchain stuff
        chat_model_run = Run(
            id=run_id,
            parent_run_id=parent_run_id,
            serialized=serialized,
            inputs={
                "messages": [[msg.model_dump() for msg in batch] for batch in messages]
            },
            extra=kwargs,
            events=[{"name": "start", "time": start_time}],
            start_time=start_time,
            run_type="llm",
            tags=tags,
            name=name,  # type: ignore[arg-type]
        )

        self._start_trace(chat_model_run)
        self._on_chat_model_start(chat_model_run)
        return chat_model_run

    def _on_chat_model_start(self, run: Run) -> None:
        """Process the Chat Model Run upon start."""
        if self._skip_tracking():
            return

        self._process_start_span(run)

    def _on_llm_end(self, run: Run) -> None:
        """Process the LLM Run."""
        if self._skip_tracking():
            return

        self._process_end_span(run)

    def _on_llm_error(self, run: Run) -> None:
        """Process the LLM Run upon error."""
        if self._skip_tracking():
            return

        self._process_end_span_with_error(run)

    def _on_chain_start(self, run: Run) -> None:
        """Process the Chain Run upon start."""
        if self._skip_tracking():
            return

        self._process_start_span(run)

    def _on_chain_end(self, run: Run) -> None:
        """Process the Chain Run."""
        if self._skip_tracking():
            return

        self._process_end_span(run)

    def _on_chain_error(self, run: Run) -> None:
        """Process the Chain Run upon error."""
        if self._skip_tracking():
            return

        self._process_end_span_with_error(run)

    def _on_tool_start(self, run: Run) -> None:
        """Process the Tool Run upon start."""
        if self._skip_tracking():
            return

        self._process_start_span(run)

    def _on_tool_end(self, run: Run) -> None:
        """Process the Tool Run."""
        if self._skip_tracking():
            return

        self._process_end_span(run)

    def _on_tool_error(self, run: Run) -> None:
        """Process the Tool Run upon error."""
        if self._skip_tracking():
            return

        self._process_end_span_with_error(run)



# This file was auto-generated by Fern from our API Definition.

import datetime as dt
import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .webhook_event_object_event_type import WebhookEventObjectEventType


class WebhookEventObject(UniversalBaseModel):
    id: str
    event_type: WebhookEventObjectEventType
    alert_id: str
    workspace_id: str
    user_name: str
    payload: typing.Dict[str, typing.Optional[typing.Any]]
    created_at: dt.datetime
    max_retries: typing.Optional[int] = None
    url: str
    secret: typing.Optional[str] = None
    headers: typing.Optional[typing.Dict[str, str]] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



import logging
import crewai
import functools
from opik.decorator import tracker as opik_tracker
from typing import Optional

LOGGER = logging.getLogger(__name__)


def patch_flow_init(project_name: Optional[str] = None) -> None:
    original_init = crewai.Flow.__init__
    if hasattr(patch_flow_init, "_patched"):
        return

    @functools.wraps(original_init)
    def _init_wrapper(self, *args, **kwargs) -> None:  # type: ignore
        original_init(self, *args, **kwargs)

        try:
            flow_registered_methods = getattr(self, "_methods", {})
            for method_name, method in list(flow_registered_methods.items()):
                if getattr(method, "opik_tracked", False):
                    continue

                decorated = opik_tracker.track(
                    project_name=project_name,
                    tags=["crewai"],
                    metadata={"created_from": "crewai"},
                )(method)

                flow_registered_methods[method_name] = decorated
        except Exception:
            LOGGER.error(
                "An error occurred during Opik instrumentation of CrewAI Flow",
                exc_info=True,
            )

    crewai.Flow.__init__ = _init_wrapper  # type: ignore[assignment]

    setattr(patch_flow_init, "_patched", True)  # type: ignore[attr-defined]


def patch_flow_kickoff_async(project_name: Optional[str] = None) -> None:
    # We only need to patch the async version of the kickoff method because
    # the sync version calls it internally
    original_kickoff_async = crewai.Flow.kickoff_async
    if hasattr(patch_flow_kickoff_async, "_patched"):
        return

    @functools.wraps(original_kickoff_async)
    async def _kickoff_async_wrapper(self, *args, **kwargs):  # type: ignore
        wrapped = opik_tracker.track(
            project_name=project_name,
            tags=["crewai"],
            name="Flow.kickoff_async",
            metadata={"created_from": "crewai"},
        )(original_kickoff_async)
        return await wrapped(self, *args, **kwargs)

    crewai.Flow.kickoff_async = _kickoff_async_wrapper  # type: ignore[assignment]

    setattr(patch_flow_kickoff_async, "_patched", True)  # type: ignore[attr-defined]



import abc
import logging
from typing import Any, Callable, Dict, Type
import pydantic
import tenacity

import opik.logging_messages as logging_messages
import opik.exceptions as exceptions
from . import messages
from ..jsonable_encoder import encode
from .. import dict_utils
from ..rate_limit import rate_limit
from ..rest_api.types import (
    feedback_score_batch_item,
    guardrail,
    feedback_score_batch_item_thread,
)
from ..rest_api import core as rest_api_core
from ..rest_api import client as rest_api_client

LOGGER = logging.getLogger(__name__)


MessageProcessingHandler = Callable[[messages.BaseMessage], None]


class BaseMessageProcessor(abc.ABC):
    @abc.abstractmethod
    def process(
        self,
        message: messages.BaseMessage,
    ) -> None:
        pass


class OpikMessageProcessor(BaseMessageProcessor):
    def __init__(
        self, rest_client: rest_api_client.OpikApi, batch_memory_limit_mb: int = 50
    ):
        self._rest_client = rest_client
        self._batch_memory_limit_mb = batch_memory_limit_mb

        self._handlers: Dict[Type, MessageProcessingHandler] = {
            messages.CreateSpanMessage: self._process_create_span_message,  # type: ignore
            messages.CreateTraceMessage: self._process_create_trace_message,  # type: ignore
            messages.UpdateSpanMessage: self._process_update_span_message,  # type: ignore
            messages.UpdateTraceMessage: self._process_update_trace_message,  # type: ignore
            messages.AddTraceFeedbackScoresBatchMessage: self._process_add_trace_feedback_scores_batch_message,  # type: ignore
            messages.AddSpanFeedbackScoresBatchMessage: self._process_add_span_feedback_scores_batch_message,  # type: ignore
            messages.AddThreadsFeedbackScoresBatchMessage: self._process_add_threads_feedback_scores_batch_message,  # type: ignore
            messages.CreateSpansBatchMessage: self._process_create_spans_batch_message,  # type: ignore
            messages.CreateTraceBatchMessage: self._process_create_traces_batch_message,  # type: ignore
            messages.GuardrailBatchMessage: self._process_guardrail_batch_message,  # type: ignore
        }

    def process(self, message: messages.BaseMessage) -> None:
        message_type = type(message)
        handler = self._handlers.get(message_type)
        if handler is None:
            LOGGER.debug("Unknown type of message - %s", message_type.__name__)
            return

        try:
            handler(message)
        except rest_api_core.ApiError as exception:
            if exception.status_code == 409:
                # sometimes a retry mechanism works in a way that it sends the same request 2 times.
                # if the backend rejects the second request, we don't want users to see an error.
                return
            elif exception.status_code == 429:
                if exception.headers is not None:
                    rate_limiter = rate_limit.parse_rate_limit(exception.headers)
                    if rate_limiter is not None:
                        raise exceptions.OpikCloudRequestsRateLimited(
                            headers=exception.headers,
                            retry_after=rate_limiter.retry_after(),
                        )

            error_tracking_extra = _generate_error_tracking_extra(exception, message)
            LOGGER.error(
                logging_messages.FAILED_TO_PROCESS_MESSAGE_IN_BACKGROUND_STREAMER,
                message_type.__name__,
                str(exception),
                extra={"error_tracking_extra": error_tracking_extra},
            )
        except tenacity.RetryError as retry_error:
            cause = retry_error.last_attempt.exception()
            error_tracking_extra = _generate_error_tracking_extra(cause, message)
            LOGGER.error(
                logging_messages.FAILED_TO_PROCESS_MESSAGE_IN_BACKGROUND_STREAMER,
                message_type.__name__,
                f"{cause.__class__.__name__} - {cause}",
                extra={"error_tracking_extra": error_tracking_extra},
            )
            LOGGER.warning(logging_messages.MAKE_SURE_OPIK_IS_CONFIGURED_CORRECTLY)
        except pydantic.ValidationError as validation_error:
            error_tracking_extra = _generate_error_tracking_extra(
                validation_error, message
            )
            LOGGER.error(
                "Failed to process message: '%s' due to input data validation error:\n%s\n",
                message_type.__name__,
                validation_error,
                exc_info=True,
                extra={"error_tracking_extra": error_tracking_extra},
            )
        except Exception as exception:
            error_tracking_extra = _generate_error_tracking_extra(exception, message)
            LOGGER.error(
                logging_messages.FAILED_TO_PROCESS_MESSAGE_IN_BACKGROUND_STREAMER,
                message_type.__name__,
                str(exception),
                exc_info=True,
                extra={"error_tracking_extra": error_tracking_extra},
            )
            LOGGER.warning(logging_messages.MAKE_SURE_OPIK_IS_CONFIGURED_CORRECTLY)

    def _process_create_span_message(
        self,
        message: messages.CreateSpanMessage,
    ) -> None:
        create_span_kwargs = message.as_payload_dict()
        cleaned_create_span_kwargs = dict_utils.remove_none_from_dict(
            create_span_kwargs
        )
        cleaned_create_span_kwargs = encode(cleaned_create_span_kwargs)
        LOGGER.debug("Create span request: %s", cleaned_create_span_kwargs)
        self._rest_client.spans.create_span(**cleaned_create_span_kwargs)

    def _process_create_trace_message(
        self,
        message: messages.CreateTraceMessage,
    ) -> None:
        create_trace_kwargs = message.as_payload_dict()
        cleaned_create_trace_kwargs = dict_utils.remove_none_from_dict(
            create_trace_kwargs
        )
        cleaned_create_trace_kwargs = encode(cleaned_create_trace_kwargs)
        LOGGER.debug("Create trace request: %s", cleaned_create_trace_kwargs)
        self._rest_client.traces.create_trace(**cleaned_create_trace_kwargs)

    def _process_update_span_message(
        self,
        message: messages.UpdateSpanMessage,
    ) -> None:
        update_span_kwargs = message.as_payload_dict()

        cleaned_update_span_kwargs = dict_utils.remove_none_from_dict(
            update_span_kwargs
        )
        cleaned_update_span_kwargs = encode(cleaned_update_span_kwargs)
        LOGGER.debug("Update span request: %s", cleaned_update_span_kwargs)
        self._rest_client.spans.update_span(**cleaned_update_span_kwargs)

    def _process_update_trace_message(
        self,
        message: messages.UpdateTraceMessage,
    ) -> None:
        update_trace_kwargs = message.as_payload_dict()

        cleaned_update_trace_kwargs = dict_utils.remove_none_from_dict(
            update_trace_kwargs
        )
        cleaned_update_trace_kwargs = encode(cleaned_update_trace_kwargs)
        LOGGER.debug("Update trace request: %s", cleaned_update_trace_kwargs)
        self._rest_client.traces.update_trace(**cleaned_update_trace_kwargs)
        LOGGER.debug("Sent trace %s", message.trace_id)

    def _process_add_span_feedback_scores_batch_message(
        self,
        message: messages.AddSpanFeedbackScoresBatchMessage,
    ) -> None:
        scores = [
            feedback_score_batch_item.FeedbackScoreBatchItem(**score_message.__dict__)
            for score_message in message.batch
        ]

        LOGGER.debug("Add spans feedbacks scores request of size: %d", len(scores))

        self._rest_client.spans.score_batch_of_spans(
            scores=scores,
        )
        LOGGER.debug("Sent batch of spans feedback scores %d", len(scores))

    def _process_add_trace_feedback_scores_batch_message(
        self,
        message: messages.AddTraceFeedbackScoresBatchMessage,
    ) -> None:
        scores = [
            feedback_score_batch_item.FeedbackScoreBatchItem(**score_message.__dict__)
            for score_message in message.batch
        ]

        LOGGER.debug("Add traces feedbacks scores request: %d", len(scores))

        self._rest_client.traces.score_batch_of_traces(
            scores=scores,
        )
        LOGGER.debug("Sent batch of traces feedbacks scores of size %d", len(scores))

    def _process_add_threads_feedback_scores_batch_message(
        self,
        message: messages.AddThreadsFeedbackScoresBatchMessage,
    ) -> None:
        scores = [
            feedback_score_batch_item_thread.FeedbackScoreBatchItemThread(
                **score_message.as_payload_dict()
            )
            for score_message in message.batch
        ]

        try:
            LOGGER.debug("Add threads feedbacks scores request of size %d", len(scores))
            self._rest_client.traces.score_batch_of_threads(
                scores=scores,
            )
            LOGGER.debug(
                "Sent batch of threads feedbacks scores of size %d", len(scores)
            )
        except rest_api_core.ApiError as exception:
            # In the case of AddThreadsFeedbackScoresBatchMessage, the backend will reject the request
            # if thread is not closed which can happen if the user is unaware of this fact.
            # Thus, we display the warning message.
            if exception.status_code == 409:
                LOGGER.warning(
                    "Threads feedbacks scores batch was rejected by the backend, reason: '%s'",
                    exception.body,
                )
            # propagate further to be handled in a unified error handler
            raise exception

    def _process_create_spans_batch_message(
        self, message: messages.CreateSpansBatchMessage
    ) -> None:
        LOGGER.debug("Create spans batch request of size %d", len(message.batch))
        self._rest_client.spans.create_spans(spans=message.batch)
        LOGGER.debug("Sent spans batch of size %d", len(message.batch))

    def _process_create_traces_batch_message(
        self, message: messages.CreateTraceBatchMessage
    ) -> None:
        LOGGER.debug("Create trace batch request of size %d", len(message.batch))
        self._rest_client.traces.create_traces(traces=message.batch)
        LOGGER.debug("Sent trace batch of size %d", len(message.batch))

    def _process_guardrail_batch_message(
        self,
        message: messages.GuardrailBatchMessage,
    ) -> None:
        batch = []

        for message_item in message.batch:
            guardrail_batch_item_message = guardrail.Guardrail(**message_item.__dict__)
            batch.append(guardrail_batch_item_message)

        self._rest_client.guardrails.create_guardrails(guardrails=batch)


def _generate_error_tracking_extra(
    exception: Exception, message: messages.BaseMessage
) -> Dict[str, Any]:
    result: Dict[str, Any] = {"exception": exception}

    if isinstance(exception, rest_api_core.ApiError):
        fingerprint = [type(message).__name__, type(exception).__name__]
        fingerprint.append(str(exception.status_code))
        result["fingerprint"] = fingerprint
        result["status_code"] = exception.status_code

    return result
